#include "Grid.h"
#include <vector>
#include <string>
#include <stdio.h>
#include <string.h>

using std::vector;
using std::string;

GridInfo::GridInfo() :
    m_time(0),
    m_score(0),
    m_state(State_Running),
    m_water(0),
    m_underwater(0),
    m_lambdaCollected(0),
    m_razors(0),
    m_robotPos(),
    m_fallingRocks(),
    m_growingBeards()
{
}

GridInfo::~GridInfo()
{
}

Grid::Grid() :
    m_pInfo(new GridInfo),
    m_width(0),
    m_height(0),
    m_data(0),
    m_liftPos(),
    m_flooding(0),
    m_waterproof(10),
    m_lambdaTotal(0),
    m_growth(25)
{
}

Grid::~Grid()
{
    if (m_data) {
        delete[] m_data;
    }
}

void Grid::resize(uint32_t width, uint32_t height)
{
    m_width = width;
    m_height = height;
    if (m_data) {
        delete[] m_data;
    }

    m_data = new CellState[(width+2) * (height+2)];

    for (uint32_t y = 0; y <= m_height+1; y++) {
        for (uint32_t x = 0; x <= m_width+1; x++) {
            if (x == 0 || y == 0 ||
                x == m_width+1 || y == m_height+1) {
                cell(x, y) = CS_Wall;
            } else {
                cell(x, y) = CS_Empty;
            }
        }
    }
}

uint32_t parseVal(const string& str)
{
    uint32_t val = 0;
    size_t len = str.size();
    bool started = false;
    for (size_t i = 0; i < len; i++) {
        char ch = str[i];
        if (!started && ch == ' ') {
            continue;
        }
        started = true;
        if (ch >= '0' && ch <= '9') {
            val *= 10;
            val += (ch - '0');
        } else {
            break;
        }
    }
    return val;
}

bool readGrid(FILE* f, Grid* pGrid)
{
    Grid& grid = *pGrid;

    vector<string> lines;
    bool gotCR = false;
    bool newLine = true;

    char buf[1024];
    while (!feof(f)) {
        size_t bytesRead = fread(buf, 1, 1024, f);
        for (size_t i = 0; i < bytesRead; i++) {
            char ch = buf[i];

            if (gotCR && ch == '\n') {
                gotCR = false;
                continue;
            }

            if (newLine) {
                lines.resize(lines.size() + 1);
                newLine = false;
            }

            string& line = lines.back();

            switch (ch) {
            case '\r':
                gotCR = true;
                newLine = true;
                break;
            case '\n':
                newLine = true;
                break;
            default:
                line.push_back(ch);
            }
        }
    }

    bool foundBlank = false;
    uint32_t blankIdx = (uint32_t)lines.size();
    uint32_t maxLen = 0;
    for (size_t i = 0; i < lines.size(); i++) {
        string& line = lines[i];
        uint32_t len = (uint32_t)line.size();
        if (!foundBlank) {
            if (len == 0) {
                foundBlank = true;
                blankIdx = (uint32_t)i;
            } else {
                if (len > maxLen) {
                    maxLen = len;
                }
            }
        } else {
            break;
        }
    }

    uint32_t width = maxLen;
    uint32_t height = blankIdx;
    grid.resize(width, height);

    for (uint32_t y = 1; y <= height; y++) {
        string& line = lines[height-y];
        for (uint32_t x = 1; x <= width; x++) {
            CellState cs = CS_Empty;
            if (x-1 < line.size()) {
                char ch = line[x-1];
                switch (ch) {
                case ' ': cs = CS_Empty; break;
                case '#': cs = CS_Wall; break;
                case '*': cs = CS_LORock; break;
                case '@': cs = CS_HORock; break;
                case '.': cs = CS_Earth; break;
                case '\\': cs = CS_Lambda; break;
                case 'L': cs = CS_Lift; break;
                case 'R': cs = CS_Robot; break;
                case 'A':
                case 'B':
                case 'C':
                case 'D':
                case 'E':
                case 'F':
                case 'G':
                case 'H':
                case 'I':
                    {
                        cs = CS_Trampoline(ch - 'A' + 1);
                        grid.m_trampolines[ch - 'A'].m_pos = Pos(x, y);
                        break;
                    }
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    {
                        cs = CS_Target(ch - '1' + 1);
                        grid.m_targets[ch - '1'].m_pos = Pos(x, y);
                        break;
                    }
                case 'W': cs = CS_Beard; break;
                case '!': cs = CS_Razor; break;
                default:
                    fprintf(stderr, "x=%u, y=%u, c='%c'\n",
                            x, y, ch);
                    return false;
                }
            } else {
                cs = CS_Empty;
            }
            if (csIsRobot(cs)) {
                grid.m_pInfo->m_robotPos = Pos(x, y);
            }
            if (csIsLift(cs)) {
                grid.m_liftPos = Pos(x, y);
            }
            if (csIsLambda(cs) || csIsHORock(cs)) {
                grid.m_lambdaTotal++;
            }
            grid.cell(x, y) = cs;
        }
    }

    for (uint32_t y = 1; y <= height; y++) {
        for (uint32_t x = 1; x <= width; x++) {
            if (csIsRock(grid.cell(x, y))) {
                CellState csUnder = grid.cell(x, y-1);
                if (csIsEmpty(csUnder) ||
                    (csIsRock(csUnder) &&
                     ((csIsEmpty(grid.cell(x+1, y)) &&
                       csIsEmpty(grid.cell(x+1, y-1))) ||
                      (csIsEmpty(grid.cell(x-1, y)) &&
                       csIsEmpty(grid.cell(x-1, y-1))))) ||
                    (csIsLambda(csUnder) &&
                     csIsEmpty(grid.cell(x+1, y)) &&
                     csIsEmpty(grid.cell(x+1, y-1)))) {
                    grid.m_pInfo->m_fallingRocks.insert(Pos(x, y));
                }
            } else if (csIsBeard(grid.cell(x, y))) {
                if (csIsEmpty(grid.cell(x-1, y-1)) ||
                    csIsEmpty(grid.cell(x, y-1)) ||
                    csIsEmpty(grid.cell(x+1, y-1)) ||
                    csIsEmpty(grid.cell(x-1, y)) ||
                    csIsEmpty(grid.cell(x+1, y)) ||
                    csIsEmpty(grid.cell(x-1, y+1)) ||
                    csIsEmpty(grid.cell(x, y+1)) ||
                    csIsEmpty(grid.cell(x+1, y+1))) {
                    grid.m_pInfo->m_growingBeards.insert(Pos(x, y));
                }
            }
        }
    }

    for (size_t i = blankIdx; i < lines.size(); i++) {
        string& line = lines[i];

        if (line.size() > 6 &&
            line.substr(0, 6).compare("Water ") == 0) {
            uint32_t val = parseVal(line.substr(6));
            grid.m_pInfo->m_water = val;
        }

        if (line.size() > 9 &&
            line.substr(0, 9).compare("Flooding ") == 0) {
            uint32_t val = parseVal(line.substr(9));
            grid.m_flooding = val;
        }

        if (line.size() > 11 &&
            line.substr(0, 11).compare("Waterproof ") == 0) {
            uint32_t val = parseVal(line.substr(11));
            grid.m_waterproof = val;
        }

        if (line.size() > 7 &&
            line.substr(0, 7).compare("Growth ") == 0) {
            uint32_t val = parseVal(line.substr(7));
            grid.m_growth = val;
        }

        if (line.size() > 7 &&
            line.substr(0, 7).compare("Razors ") == 0) {
            uint32_t val = parseVal(line.substr(7));
            grid.m_pInfo->m_razors = val;
        }

        if (line.size() >= 22 &&
            line.substr(0, 11).compare("Trampoline ") == 0 &&
            line.substr(12, 9).compare(" targets ") == 0 &&
            line[11] >= 'A' && line[11] <= 'I' &&
            line[21] >= '1' && line[21] <= '9') {
            grid.m_trampolines[line[11] - 'A'].m_target =
                line[21] - '1' + 1;
        }
    }

    return true;
}

void writeGrid(FILE* f, const Grid& grid, bool ansi)
{
    for (uint32_t y = grid.m_height; y >= 1; y--) {
        for (uint32_t x = 1; x <= grid.m_width; x++) {
            CellState cs = grid.cell(x, y);
            bool water = (y <= grid.m_pInfo->m_water);
            if (csIsEmpty(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;31;44m" : "\x1b[0;31m", f);
                }
                fputc(' ', f);
            } else if (csIsWall(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;31;44m" : "\x1b[0;31m", f);
                }
                fputc('#', f);
            } else if (csIsLORock(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;44m" : "\x1b[0;1m", f);
                }
                fputc('*', f);
            } else if (csIsHORock(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;35;44m" : "\x1b[0;1;35m", f);
                }
                fputc('@', f);
            } else if (csIsEarth(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;44m" : "\x1b[0m", f);
                }
                fputc('.', f);
            } else if (csIsLambda(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;33;44m" : "\x1b[0;1;33m", f);
                }
                fputc('\\', f);
            } else if (csIsLift(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;34;44m" : "\x1b[0;1;34m", f);
                }
                if (grid.m_pInfo->m_lambdaCollected == grid.m_lambdaTotal) {
                    fputc('O', f);
                } else {
                    fputc('L', f);
                }
            } else if (csIsRobot(cs)) {
                if (ansi) {
                    if (grid.m_pInfo->m_state == State_Broken) {
                        fputs(water ? "\x1b[0;1;31;44m" : "\x1b[0;1;31m", f);
                    } else if (grid.m_pInfo->m_state == State_Finished) {
                        fputs(water ? "\x1b[0;1;33;44m" : "\x1b[0;1;33m", f);
                    } else {
                        fputs(water ? "\x1b[0;1;32;44m" : "\x1b[0;1;32m", f);
                    }
                }
                fputc('R', f);
            } else if (csIsTrampoline(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;36;44m" : "\x1b[0;1;36m", f);
                }
                fputc('A' + csGetNumber(cs) - 1, f);
            } else if (csIsTarget(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;36;44m" : "\x1b[0;1;36m", f);
                }
                fputc('1' + csGetNumber(cs) - 1, f);
            } else if (csIsBeard(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;31;44m" : "\x1b[0;1;31m", f);
                }
                fputc('W', f);
            } else if (csIsRazor(cs)) {
                if (ansi) {
                    fputs(water ? "\x1b[0;1;44m" : "\x1b[0;1m", f);
                }
                fputc('!', f);
            } else {
                if (ansi) {
                    fputs(water ? "\x1b[0;44m" : "\x1b[0m", f);
                }
                fputc('?', f);
            }
        }
        if (ansi) {
            fputs("\x1b[0m", f);
        }
        fputc('\n', f);
    }

    fprintf(f, "\n");

    switch (grid.m_pInfo->m_state) {
    case State_Running: fprintf(f, "Running \n"); break;
    case State_Aborted: fprintf(f, "Aborted \n"); break;
    case State_Finished: fprintf(f, "Finished\n"); break;
    case State_Broken: fprintf(f, "Broken  \n"); break;
    }

    fprintf(f, "Trampolines");
    for (uint32_t i = 1; i <= 9; i++) {
        if (grid.m_trampolines[i-1].m_target != 0) {
            fprintf(f, " %c->%c",
                    'A' + i - 1,
                    '1' + grid.m_trampolines[i-1].m_target - 1);
        }
    }
    fprintf(f, "\n");

    fprintf(f, "Time %u\n", grid.m_pInfo->m_time);
    fprintf(f, "Score %d\n", grid.m_pInfo->m_score);
    fprintf(f, "Water %u\n", grid.m_pInfo->m_water);
    fprintf(f, "Flooding %u\n", grid.m_flooding);
    fprintf(f, "Waterproof %u\n", grid.m_waterproof);
    fprintf(f, "Underwater %u\n", grid.m_pInfo->m_underwater);
    fprintf(f, "Growth %u\n", grid.m_growth);
    fprintf(f, "Razors %u\n", grid.m_pInfo->m_razors);
}
